import importlib
import inspect
from typing import Any, Dict

from django.apps import apps
from django.http import HttpRequest, HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.generic.base import View

from hx_requests.utils import deserialize_kwargs, is_htmx_request


@method_decorator(ensure_csrf_cookie, name="dispatch")
class HtmxViewMixin(View):
    """
    Mixin to be added to views that are using HXRequests.
    Hijacks the get and post to route them to the proper
    HXRequest.
    """

    hx_requests: Dict = []

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if is_htmx_request(request):
            hx_request = self._setup_hx_request(request, *args, **kwargs)
            kwargs = self.get_extra_kwargs(request)
            return hx_request.get(request, *args, **kwargs)
        return super().get(request, *args, **kwargs)

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if is_htmx_request(request):
            hx_request = self._setup_hx_request(request, *args, **kwargs)
            kwargs = self.get_extra_kwargs(request)
            return hx_request.post(request, *args, **kwargs)
        return super().post(request, *args, **kwargs)

    def get_hx_request(self, request):
        hx_request_name = request.GET.get("hx_request_name")
        self._get_hx_request_classes()
        try:
            hx_request = next(
                hx_request
                for name, hx_request in self.hx_requests.items()
                if name == hx_request_name
            )
        except StopIteration:
            raise Exception(
                f"No HXRequest found with the name {hx_request_name}. Are you sure it's spelled correctly?"
            )
        return hx_request()

    @classmethod
    def _get_hx_request_classes(cls):
        from .hx_requests import BaseHXRequest

        # If the hx_requests are already set don't need to do the whole collection.
        if getattr(cls, "hx_requests", None):
            return
        modules = []
        hx_request_classes = {}
        for app in apps.get_app_configs():
            try:
                modules.append(importlib.import_module(f"{app.label}.hx_requests"))
            except ModuleNotFoundError:
                pass
        for module in modules:
            clsmembers = inspect.getmembers(module, inspect.isclass)
            for _, obj in clsmembers:
                if issubclass(obj, BaseHXRequest) and getattr(obj, "name", None):
                    hx_request_classes[obj.name] = obj
        cls.hx_requests = hx_request_classes

    def get_extra_kwargs(self, request):
        kwargs = {}
        for key in request.GET:
            kwargs[key] = request.GET.get(key)

        return deserialize_kwargs(**kwargs)

    def _setup_hx_request(self, request, *args, **kwargs):
        # Call get even on post to setup neccessary parts for the context
        super().get(request, *args, **kwargs)
        hx_request = self.get_hx_request(request)
        hx_request.view = self
        hx_request.setup_hx_request(request)
        return hx_request



from urllib.parse import quote_plus

from django.apps import apps
from django.db import models


def is_htmx_request(request):
    return "HX-Request" in request.headers


def serialize_kwargs(**kwargs):
    for k, v in kwargs.items():
        if isinstance(v, models.Model):
            kwargs[k] = f"___{v._meta.app_label}_{v._meta.model.__name__}_{v.pk}"
    return kwargs


def deserialize_kwargs(**kwargs):
    for k, v in kwargs.items():
        if v.startswith("___"):
            app_label, model, pk = v.replace("___", "").split("_")
            model = apps.get_model(app_label, model)
            instance = model.objects.get(pk=pk)
            kwargs[k] = instance
    return kwargs


def get_url(context, hx_request_name, object, **kwargs):
    url = context["request"].path
    if not "?" in url:
        url += f"?hx_request_name={hx_request_name}"
    else:
        url += f"&hx_request_name={hx_request_name}"
    url += (
        f"&object={object._meta.app_label}_{object._meta.model.__name__}_{object.pk}"
        if object
        else ""
    )
    serialized_kwargs = serialize_kwargs(**kwargs)
    extra_params = ""
    for k, v in serialized_kwargs.items():
        extra_params += f"&{k}={quote_plus(str(v))}"
    url += extra_params
    return url


def get_csrf_token(context):
    token = (
        context["request"].headers.get("cookie").split("csrftoken=")[1].split(";")[0]
    )
    return token
